"""
Wisconsin Autonomous - https://www.wisconsinautonomous.org

Copyright (c) 2021 wisconsinautonomous.org
All rights reserved.

Use of this source code is governed by a BSD-style license that can be found
in the LICENSE file at the top level of the repo
"""

# WA Simulator
from wa_simulator.core import WAVector, WAQuaternion
from wa_simulator.utils import _check_field, get_wa_data_directory

# Chrono specific imports
import pychrono as chrono
import pychrono.vehicle as veh

# ----------------------
# Data loading utilities
# ----------------------

import pathlib
import contextlib


def _update_chrono_data_directories():
    global _DATA_DIRECTORY, _CHRONO_DATA_DIRECTORY, _CHRONO_VEH_DATA_DIRECTORY

    if not _OVERRIDE_CHRONO_DIRECTORIES and _DATA_DIRECTORY != get_wa_data_directory():
        _DATA_DIRECTORY = get_wa_data_directory()

        _CHRONO_DATA_DIRECTORY = str(pathlib.Path(_DATA_DIRECTORY) / "chrono" / " ")[:-1]
        _CHRONO_VEH_DATA_DIRECTORY = str(pathlib.Path(_DATA_DIRECTORY) / "chrono" / "vehicle" / " ")[:-1]

        # Update the chrono paths
        chrono.SetChronoDataPath(_CHRONO_DATA_DIRECTORY)
        veh.SetDataPath(_CHRONO_VEH_DATA_DIRECTORY)


def set_chrono_directories_override(override: bool):
    """Set the chrono data directories to have the passed override value.

    Normally, it is assumed that the chrono data directories are located inside wherever :meth:`~get_wa_data_directory`
    is set to. This method will change the override functionality to either continue to look in this location, or
    allow the user to override this. To override the actual directories, use :meth:`~set_chrono_data_directory` and 
    :meth:`~set_chrono_vehicle_data_directory`.

    If set to True, the current paths will remain the same, but will not change without either setting them using the
    setters or when the wa data directory is changed (which would be the case if set to False).

    Args:
        bool: Override if True, don't if False
    """
    global _OVERRIDE_CHRONO_DIRECTORIES

    _OVERRIDE_CHRONO_DIRECTORIES = override


def set_chrono_data_directory(path: str):
    """Set the chrono data directory

    Normally, it is assumed that the chrono data directory is located inside wherever :meth:`~get_wa_data_directory`
    is set to. Using this method will override that functionality and statically assign the chrono data directory.
    If you'd like to return to the overriding functionality, call :meth:`~set_chrono_data_directory_override`.

    Args:
        path (str): relative (or absolute) path where the data is stored
    """
    global _CHRONO_DATA_DIRECTORY, _OVERRIDE_CHRONO_DIRECTORIES

    _CHRONO_DATA_DIRECTORY = path
    _OVERRIDE_CHRONO_DIRECTORIES = True

    chrono.SetChronoDataPath(_CHRONO_DATA_DIRECTORY)


def set_chrono_vehicle_data_directory(path: str):
    """Set the chrono vehicle data directory

    Normally, it is assumed that the chrono vehicle data directory is located inside wherever :meth:`~get_wa_data_directory`
    is set to. Using this method will override that functionality and statically assign the chrono vehicle data directory.
    If you'd like to return to the overriding functionality, call :meth:`~set_chrono_directories_override`.

    Args:
        path (str): relative (or absolute) path where the data is stored
    """
    global _CHRONO_DATA_DIRECTORY, _OVERRIDE_CHRONO_DIRECTORIES

    _CHRONO_DATA_DIRECTORY = path
    _OVERRIDE_CHRONO_DIRECTORIES = True

    veh.SetDataPath(_CHRONO_VEH_DATA_DIRECTORY)


def get_chrono_data_file(filename: str) -> str:
    """Get the absolute path for the filename passed relative to the :data:`~CHRONO_DATA_DIRECTORY`.

    .. highlight:: python
    .. code:: python

        # Example Usage
        from wa_simulator import get_chrono_data_file

        # By default, the data directory will be set to '<installation path of wa_simulator>/data/chrono'
        path = get_chrono_data_file('test.json')

        print(path) # -> '<installation path of wa_simulator>/data/chrono/test.json'

    Args:
        filename (str): file relative to the data folder to get the absolute path for

    Returns:
        str: the absolute path of the file
    """
    _update_chrono_data_directories()
    return str(pathlib.Path(_CHRONO_DATA_DIRECTORY) / filename)


def get_chrono_vehicle_data_file(filename: str) -> str:
    """Get the absolute path for the filename passed relative to the :data:`~CHRONO_VEH_DATA_DIRECTORY`.

    .. highlight:: python
    .. code:: python

        # Example Usage
        from wa_simulator import get_chrono_vehicle_data_file

        # By default, the data directory will be set to '<installation path of wa_simulator>/data/chrono/vehicle'
        path = get_chrono_vehicle_data_file('test.json')

        print(path) # -> '<installation path of wa_simulator>/data/chrono/vehicle/test.json'

    Args:
        filename (str): file relative to the data folder to get the absolute path for

    Returns:
        str: the absolute path of the file
    """
    _update_chrono_data_directories()
    return str(pathlib.Path(_CHRONO_VEH_DATA_DIRECTORY) / filename)


# Initialze the chrono data directory to in-repo data directory
_DATA_DIRECTORY = ""
_CHRONO_DATA_DIRECTORY = ""
_CHRONO_VEH_DATA_DIRECTORY = ""
_OVERRIDE_CHRONO_DIRECTORIES = False
_update_chrono_data_directories()

# ------------------
# vector conversions
# ------------------


def ChVector_to_WAVector(vector: chrono.ChVectorD):
    """Converts a ChVector to a WAVector

    Args:
        vector (ChVector): The vector to convert
    """
    return WAVector([vector.x, vector.y, vector.z])


def WAVector_to_ChVector(vector: WAVector):
    """Converts a WAVector to a ChVector

    Args:
        vector (WAVector): The vector to convert
    """
    return chrono.ChVectorD(vector.x, vector.y, vector.z)


def ChQuaternion_to_WAQuaternion(quaternion: chrono.ChQuaternionD):
    """Converts a ChQuaternion to a WAQuaternion

    Args:
        quaternion (ChQuaternion): The quaternion to convert
    """
    return WAQuaternion([quaternion.e1, quaternion.e2, quaternion.e3, quaternion.e0])


def WAQuaternion_to_ChQuaternion(quaternion: WAQuaternion):
    """Converts a WAQuaternion to a ChQuaternion

    Args:
        quaternion (WAQuaternion): The quaternion to convert
    """
    return chrono.ChQuaternionD(quaternion.x, quaternion.y, quaternion.z, quaternion.w)

# ----------------------
# json related utilities
# ----------------------


def ChFrame_from_json(j: dict):
    """Creates a ChFrame from a json object.

    Args:
        j (dict): The json object that will be converted to a ChFrame

    Returns:
        ChFrameD: The frame created from the json object
    """

    # Validate the json file
    _check_field(j, 'Position', field_type=list)
    _check_field(j, 'Rotation', field_type=list)

    # Do the conversion
    pos = j['Position']
    rot = j['Rotation']
    return chrono.ChFrameD(chrono.ChVectorD(pos[0], pos[1], pos[2]), chrono.ChQuaternionD(rot[0], rot[1], rot[2], rot[3]))


def ChVector_from_list(l: list, vector_type=chrono.ChVectorD):
    """Creates a ChVector from a list

    Args:
        l (list): The list to convert to a ChVector
        vector_type (Type): The ChVector type to convert to
    """
    if len(l) != 3:
        raise TypeError('list must have length 3')

    return vector_type(float(l[0]), float(l[1]), float(l[2]))



"""
Wisconsin Autonomous - https://www.wisconsinautonomous.org

Copyright (c) 2021 wisconsinautonomous.org
All rights reserved.

Use of this source code is governed by a BSD-style license that can be found
in the LICENSE file at the top level of the repo
"""

# WA Simulator
from wa_simulator.core import WA_PI
from wa_simulator.utils import _load_json, _check_field, _WAStaticAttribute, get_wa_data_file
from wa_simulator.environment import WAEnvironment, load_environment_from_json, WABody
from wa_simulator.chrono.utils import ChVector_to_WAVector, WAVector_to_ChVector, get_chrono_data_file

# Chrono specific imports
import pychrono as chrono
import pychrono.vehicle as veh


def load_chrono_terrain_from_json(system: 'WAChronoSystem', filename: str):
    """Load a ChTerrain from a json specification file

    Args:
        filename (str): the relative path to a terrain json file
        system (WAChronoSystem): the chrono system used to handle the terrain

    Returns:
        ChTerrain: The loaded terrain
    """
    j = _load_json(filename)

    # Validate the json file
    _check_field(j, 'Terrain', field_type=dict)

    t = j['Terrain']
    _check_field(t, 'Input File', field_type=str)
    _check_field(t, 'Texture', field_type=str, optional=True)

    terrain = veh.RigidTerrain(system._system, chrono.GetChronoDataFile(t['Input File']))  # noqa

    # Add texture to the terrain, if desired
    if 'Texture' in t:
        texture_filename = chrono.GetChronoDataFile(t['Texture'])

        vis_mat = chrono.ChVisualMaterial()
        vis_mat.SetKdTexture(texture_filename)
        vis_mat.SetSpecularColor(chrono.ChVectorF(0.0, 0.0, 0.0))
        vis_mat.SetFresnelMin(0)
        vis_mat.SetFresnelMax(0.1)

        patch_asset = terrain.GetPatches()[0].GetGroundBody().GetAssets()[0]
        patch_visual_asset = chrono.CastToChVisualization(patch_asset)
        patch_visual_asset.material_list.append(vis_mat)

    return terrain


class WAChronoEnvironment(WAEnvironment):
    """The environment wrapper that's responsible for holding Chrono assets and the terrain

    TODO: Add assets from file

    Args:
        system (WASystem): the wa system that wraps a ChSystem
        filename (str): the json specification file describing the environment
    """

    # Global filenames for environment models
    _EGP_ENV_MODEL_FILE = "environments/ev_grand_prix.json"

    EGP_ENV_MODEL_FILE = _WAStaticAttribute('_EGP_ENV_MODEL_FILE', get_chrono_data_file)

    def __init__(self, system: 'WAChronoSystem', filename: str):
        super().__init__()

        self._system = system
        self._terrain = load_chrono_terrain_from_json(system, filename)

        load_environment_from_json(self, filename)

    def synchronize(self, time):
        self._terrain.Synchronize(time)

    def advance(self, step):
        self._terrain.Advance(step)

        for asset in self._updating_assets:
            asset.chrono_body.SetPos(WAVector_to_ChVector(asset.position))

    def add_asset(self, asset: 'Any'):
        if isinstance(asset, chrono.ChBody):
            self._system._system.AddBody(asset)
            return
        if isinstance(asset, WABody):
            if not hasattr(asset, 'size') or not hasattr(asset, 'position'):
                raise AttributeError("Body must have 'size', and 'position' fields")

            position = asset.position
            yaw = 0 if not hasattr(asset, 'yaw') else asset.yaw
            size = asset.size

            body_type = 'box'
            if hasattr(asset, 'body_type'):
                body_type = asset.body_type

            if body_type == 'sphere':
                body = chrono.ChBodyEasySphere(size.length, 1000, True, False)
                body.SetBodyFixed(True)
            elif body_type == 'box':
                body = chrono.ChBodyEasyBox(size.x, size.y, size.z, 1000, True, False)
                body.SetBodyFixed(True)
            else:
                raise ValueError(f"'{asset.body_type}' not a supported body type.")

            body.SetPos(WAVector_to_ChVector(position))
            body.SetRot(chrono.Q_from_AngZ(-yaw + WA_PI / 2))

            if hasattr(asset, 'color'):
                color = asset.color
                body.AddAsset(chrono.ChColorAsset(chrono.ChColor(color.x, color.y, color.z)))

                texture = chrono.ChVisualMaterial()
                texture.SetDiffuseColor(chrono.ChVectorF(color.x, color.y, color.z))
                chrono.CastToChVisualization(body.GetAssets()[0]).material_list.append(texture)

            if hasattr(asset, 'texture'):
                texture = chrono.ChVisualMaterial()
                texture.SetKdTexture(get_wa_data_file(asset.texture))
                chrono.CastToChVisualization(body.GetAssets()[0]).material_list.append(texture)

            self._system._system.AddBody(body)

            asset.chrono_body = body

        super().add_asset(asset)



"""
Wisconsin Autonomous - https://www.wisconsinautonomous.org

Copyright (c) 2021 wisconsinautonomous.org
All rights reserved.

Use of this source code is governed by a BSD-style license that can be found
in the LICENSE file at the top level of the repo
"""

import argparse
from pyrr import Vector3, Quaternion
from numbers import Number
import math

# ---------
# Constants
# ---------

WA_GRAVITY = 9.81  # [m/s^2]
"""Gravitational constant (m/s^2)"""

WA_EARTH_RADIUS = 6371000.0  # [m]
"""Radius of the earth (m)"""

WA_PI = 3.141592653589793
"""PI"""

# ---------------------------------
# Vector/Quaternion/Math core items
# ---------------------------------


class WAVector(Vector3):
    """Vector object that contains x,y,z values

    This vector inherits directly from |Vector3|_, which is in the `Pyrr <https://pyrr.readthedocs.io/en/latest/index.html>`_ package. 
    The only purpose to having this class at all is purely semantic; most other classes in this simulator are prefixed with 
    `WA`, so it was desired to the same with this class. No additional functionality is added.

    .. highlight:: python
    .. code-block:: python

        # Example uses

        from wa_simulator import WAVector

        # Simple instantiation
        v1 = WAVector()
        v2 = WAVector([2, 1, 4])
        print(v1) # -> [0., 0., 0.]
        print(v2) # -> [2., 1., 4.]

        # Grab values
        print(v1.x) # -> 0.0
        print(v2.z) # -> 4.0
        print(v2[1]) # -> 1.0

        # Simple math
        v3 = WAVector([1., 1., 1.])
        print(v2 + v3) # -> [3., 2., 5.]
        print(v2 - v3) # -> [1., 0., 3.]

        # Dot product
        print(v2 * v3) # -> 7.0

        # Cross product
        print(v2 ^ v3) # -> [-3., 2., 1.]

        # Get the length
        print(v3.length) # -> 1.732

        # Normalize a vector
        v3.normalize()
        print(v3) # -> [.5773, .5773, .5773]
        print(v3.length) # -> 1.


    Please refer to the Pyrr documentation for a more detailed explaination of |Vector3|_.

    .. |Vector3| replace:: :code:`pyrr.Vector3`
    .. _Vector3: https://pyrr.readthedocs.io/en/latest/oo_api_vector.html#module-pyrr.objects.vector3
    """

    def __new__(cls, value=None, dtype=None):
        """Will construct a new :class:`~WAVector`, but change the type to float if no dtype is passed"""
        if dtype is not None:
            return super().__new__(cls, value, dtype)
        else:
            if value is not None:
                value = [float(v) for v in value]
            return super().__new__(cls, value, float)


class WAQuaternion(Quaternion):
    """Quaternion object that contains x,y,z,w values

    This quaternion class inherits directly from |Quaternion|_, which is in the `Pyrr <https://pyrr.readthedocs.io/en/latest/index.html>`_ package. 
    The only purpose to having this class at all is purely semantic; most other classes in this simulator are prefixed with 
    `WA`, so it was desired to the same with this class. Minimal additional functionality is added.

    Quaternions are rather uncommon outside of the robotics community, so a short explanation about their purpose is provided here.
    For additional information, please refer to 
    `this <https://www.allaboutcircuits.com/technical-articles/dont-get-lost-in-deep-space-understanding-quaternions/>`_ 
    and `this <https://en.wikipedia.org/wiki/Quaternion>`_.

    Quaternions are essentially mathematical operators that rotate and scale vectors. In 2D dimensional space, it's fairly likely
    you've used complex numbers to rotate vectors. For example, you may know that if you multiply a vector by the complex
    number :code:`i`, you will rotate that vector by 90 degrees. Furthermore, in 2D, to rotate any vector, you need two values: one real
    and one complex. This is, in a way, very similar to quaternions. In essence, 
    to provide the ability to rotate a vector in any desired way, you actually need four values. This is a bit counterintuitive,
    but the reasoning is outside the scope of this description (see `this video <https://www.youtube.com/watch?v=3BR8tK-LuB0>`_ 
    if you're interested). As opposed to a rotation vector in 2D, the 3D rotation vector, a quaternion, has one real and three complex
    values. These complex values, of which you've probably seen before, are named i, j and k. The real value is commonly refered to
    as w. To now rotate a 3D vector in any direction, a quaternion and it's conjugate (the original quaternion but with the opposite 
    rotation) is applied. Quaternion's rotations can be accumulated by simply "crossing" their values.

    Many have heard or used `Euler angles <https://en.wikipedia.org/wiki/Euler_angles>`_ instead. These are far more intuitive
    than quaternions, but are prone to "bugs", like `gimbal lock <https://en.wikipedia.org/wiki/Gimbal_lock>`_.

    .. highlight:: python
    .. code-block:: python

        # Example uses

        from wa_simulator import WAQuaternion, WAVector, WA_PI

        # Simple instantiation
        q1 = WAQuaternion()
        q2 = WAQuaternion([2, 1, 4, 1])
        q3 = WAQuaternion.from_z_rotation(WA_PI / 2)
        print(q1) # -> [0., 0., 0., 1.]
        print(q2) # -> [2., 1., 4., 1.]
        print(q3) # -> [0., 0., 0.70711, 0.70711]

        # Grab values
        print(q1.x) # -> 0.0
        print(q2.z) # -> 4.0
        print(q3[2]) # -> 0.70711

        # Accumulating quaternion rotations
        q4 = WAQuaternion.from_x_rotation(WA_PI)
        q5 = q3 * q4 # represents 90 degree rotation about the z, then 180 degree about the x
        print(q5) # -> [0.70711, 0.70711, 0., 0.]

        # Rotate a vector
        v = WAVector([0, 1, 0])
        print(q5 * v) # -> [1., 0., 0.]

        # Undo a rotation
        v2 = q5 * v
        print(~q5 * v2) # -> [0., 1., 0.]

        # Grab the yaw
        print(q3.to_euler_yaw()) # -> 1.57

    Please refer to the Pyrr documentation for a more detailed explaination of |Quaternion|_.

    .. |Quaternion| replace:: :code:`pyrr.Quaternion`
    .. _Quaternion: https://pyrr.readthedocs.io/en/latest/oo_api_quaternion.html#id1
    """
    def __new__(cls, value=None, dtype=None):
        """Will construct a new :class:`~WAVector`, but change the type to float if no dtype is passed"""
        if dtype is not None:
            return super().__new__(cls, value, dtype)
        else:
            if value is not None:
                value = [float(v) for v in value]
            return super().__new__(cls, value, float)

    def to_euler_roll(self) -> float:
        """Converts this quaternion to euler angles and returns the roll value (see `here <https://en.wikipedia.org/wiki/Conversion_between_quaternions_and_Euler_angles>`_).

        Returns:
            float: roll
        """
        sinr_cosp = 2 * (self.w * self.x + self.y * self.z)
        cosr_cosp = 1 - 2 * (self.x * self.x + self.y * self.y)
        roll = math.atan2(sinr_cosp, cosr_cosp)

        return roll

    def to_euler_pitch(self) -> float:
        """Converts this quaternion to euler angles and returns the pitch value (see `here <https://en.wikipedia.org/wiki/Conversion_between_quaternions_and_Euler_angles>`_)

        Returns:
            float: pitch
        """
        sinp = 2 * (self.w * self.y - self.z * self.x)
        pitch = math.copysign(math.pi / 2, sinp) if abs(sinp) >= 1 else math.asin(sinp)  # noqa

        return pitch

    def to_euler_yaw(self) -> float:
        """Converts this quaternion to euler angles and returns the yaw value (see `here <https://en.wikipedia.org/wiki/Conversion_between_quaternions_and_Euler_angles>`_)

        Returns:
            float: yaw
        """
        siny_cosp = 2 * (self.w * self.z + self.x * self.y)
        cosy_cosp = 1 - 2 * (self.z * self.z + self.y * self.y)
        yaw = math.atan2(siny_cosp, cosy_cosp)

        return yaw

    def to_euler(self) -> tuple:
        """Converts this quaternion to euler angles (see `here <https://en.wikipedia.org/wiki/Conversion_between_quaternions_and_Euler_angles>`_)

        Returns:
            tuple: (roll, pitch, yaw)
        """

        return self.to_euler_roll(), self.to_euler_pitch(), self.to_euler_yaw()


# --------------
# CLI core items
# --------------


class WAArgumentParser(argparse.ArgumentParser):
    """Argument parser wrapper.

    Has handy default arguments method

    .. highlight:: python
    .. code-block:: python

        # Example
        from wa_simulator import WAArgumentParser

        # Instantiate the parser and custom arguments
        parser = wa.WAArgumentParser(use_sim_defaults=True)
        parser.add_argument("-mv", "--matplotlib", action="store_true", help="Use matplotlib to visualize", default=False)
        parser.add_argument("-q", "--quiet", action="store_true", help="Silence any terminal output", default=False)

        # Parse the command line inputs
        args = parser.parse_args()

        # Grab the parsed options
        step_size = args.step_size # from default options
        render_step_size = args.render_step_size # from default options
        use_matplotlib = args.matplotlib # from default options
        is_quiet = args.quiet

    Args:
        use_sim_defaults (bool, optional): Use the default arguments (can add more). Defaults to True.
        skip_defaults (list, optional): The default arguments to skip.
        description (str, optional): Description used to describe the simulation. Printed by the help menu. Defaults to ("Wisconsin Autonomous Simulator").
    """

    def __init__(self, use_sim_defaults: bool = True, skip_defaults: list = [], description=("Wisconsin Autonomous Simulator")):
        super().__init__(description=description)

        if use_sim_defaults:
            self.add_default_sim_arguments(skip_defaults)

    def add_default_sim_arguments(self, skip_defaults: list = []):
        """Helper function to add predefined default arguments that can be used for most vehicle related simulations

        Set :code:`use_sim_defaults` in :code:`__init__` to :code:`True` to have this be called automatically.

        This function, unless specified in :code:`skip_defaults`, will add the following arguments:

        * -s, - -sim_step_size: Simulation step size
        * -rs, - -render_step_size: Rendering step size
        * -e, - -end_time: Simulation end time

        Args:
            skip_defaults (list, optional): The default arguments to skip.
        """

        def no_skip(s): return s not in skip_defaults

        if no_skip('s') and no_skip('step_size') and no_skip('sim_step_size'):
            self.add_argument(
                "-s",
                "--sim_step_size",
                type=float,
                help="Simulation Step Size [s]",
                default=3e-3,
                dest="step_size",
            )

        if no_skip('rs') and no_skip('render_step_size'):
            self.add_argument(
                "-rs",
                "--render_step_size",
                type=float,
                help="Render Update Rate [Hz]",
                default=1 / 10.0,
            )

        if no_skip('e') and no_skip('end_time'):
            self.add_argument(
                "-e",
                "--end_time",
                type=float,
                help="Simulation End Time [s]",
                default=120,
            )

        # if no_skip('r') and no_skip('record'):
        #     self.add_argument(
        #         "-r",
        #         "--record",
        #         action="store_true",
        #         help="Record Simple State Data",
        #         default=False,
        #     )



"""
Wisconsin Autonomous - https://www.wisconsinautonomous.org

Copyright (c) 2021 wisconsinautonomous.org
All rights reserved.

Use of this source code is governed by a BSD-style license that can be found
in the LICENSE file at the top level of the repo
"""

import unittest
import numpy as np

# Import the core module
from wa_simulator.vehicle import WALinearKinematicBicycle

# -----
# Tests
# -----


class TestWALinearKinematicBicycle(unittest.TestCase):
    """Tests method related to the WALinearKinematicBicycle"""

    pass


if __name__ == '__main__':
    unittest.main()



"""
Wisconsin Autonomous - https://www.wisconsinautonomous.org

Copyright (c) 2021 wisconsinautonomous.org
All rights reserved.

Use of this source code is governed by a BSD-style license that can be found
in the LICENSE file at the top level of the repo
"""

# WA Simulator
from wa_simulator.system import WASystem

# Chrono specific imports
import pychrono as chrono


class WAChronoSystem(WASystem):
    """Chrono system wrapper.

    Calls an update for the Chrono dynamics each :meth:`~advance` step.

    To instantiate a WAChronoSystem, you can either pass in various parameters directly, i.e. ``WAChronoSystem(step_size, render_step_size, ...)``
    or pass a ``argparse.Namespace`` in directly. This object is what is returned from :meth:`~WAArgumentParser.parse_args`.

    .. highlight:: python
    .. code:: python

        # Usage
        from wa_simulator.chrono import WAArgumentParser, WAChronoSystem

        # Passing values directly to the system
        system = WAChronoSystem(step_size=1e-3, render_step_size=2e-3, end_time=20, contact_method="SMC")

        # Or using WAArgumentParser
        parser = WAArgumentParser(use_sim_defaults=True)
        args = parser.parse_args()

        system = WAChronoSystem(args=args)

    Args:
        step_size (double): simulation step size. Defaults to 3e-3.
        render_step_size (double, optional): render step size. Defaults to 2e-2.
        end_time (float, optional): the end time for the simulation. Defaults to 120 seconds.
        args (argparse.Namespace, optional): the output namespace from argparse (see explanation above)
        contact_method (str, optional): the contact method to use (NSC or SMC). Defaults to "NSC".

    Raises:
        TypeError: verify the contact method is a string
    """

    def __init__(self, step_size: float = 3e-3, render_step_size: float = 2e-2, end_time: float = 120, contact_method: str = "NSC", args: 'argparse.Namespace' = None):
        super().__init__(step_size, render_step_size, end_time, args)

        if not isinstance(contact_method, str):
            raise TypeError("Contact method must be of type str")

        if contact_method == "NSC":
            system = chrono.ChSystemNSC()
        elif contact_method == "SMC":
            system = chrono.ChSystemSMC()
        system.Set_G_acc(chrono.ChVectorD(0, 0, -9.81))
        system.SetSolverType(chrono.ChSolver.Type_BARZILAIBORWEIN)
        system.SetSolverMaxIterations(150)
        system.SetMaxPenetrationRecoverySpeed(4.0)

        self._system = system

    def advance(self):
        """Advance the simulation.

        Will update the dynamics of the chrono simulation
        """
        self.step_number += 1
        self.time = self._system.GetChTime()

        self._system.DoStepDynamics(self.step_size)



"""A setuptools based setup module.
See:
https://packaging.python.org/guides/distributing-packages-using-setuptools/
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
from pathlib import Path
import os
import re

# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...


def read(fname):
    here = Path(__file__).parent.resolve()
    return (here / fname).read_text(encoding="utf-8")


def get_package_data(directory, rm='wa_simulator/'):
    paths = []
    for (path, directories, filenames) in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join(path.split(rm)[1], filename))
    return paths


# Get the version (borrowed from SQLAlchemy)
def get_version():
    with open(os.path.join(os.path.dirname(__file__), "wa_simulator", "_version.py")) as fp:
        return (
            re.compile(
                r""".*__version__ = ["'](.*?)['"]""", re.S).match(fp.read()).group(1)
        )


setup(
    name="wa_simulator",
    version=get_version(),
    author="Wisconsin Autonomous",
    author_email="wisconsinautonomous@studentorg.wisc.edu",
    license="BSD3",
    description=(
        "Simulation tool for prototyping autonomous vehicle related algorithms. Wrapper of the PyChrono simulator."
    ),
    long_description=read("README.md"),
    long_description_content_type="text/markdown",  # Optional (see note above)
    packages=find_packages(),
    package_data={"wa_simulator": get_package_data("wa_simulator/data")},
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Topic :: Software Development",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3",
    ],
    keywords="simulation, autonomous vehicles, robotics",
    python_requires=">=3.0",
    install_requires=[
        "matplotlib>=3.3.2",
        "numpy>=1.19.3",
        "scipy",
        "IPython",
        "pyrr",
        "dill"
    ],
    project_urls={  # Optional
        "Homepage": "https://github.com/WisconsinAutonomous/wa_simulator/",
        "Documentation": "https://WisconsinAutononomous.github.io/wa_simulator",
        "Bug Reports": "https://github.com/WisconsinAutonomous/wa_simulator/issues",
        "Source Code": "https://github.com/WisconsinAutonomous/wa_simulator/",
        "Our Team!": "https://wisconsinautonomous.org",
    },
)



"""
Requires the `PyChrono <https://projectchrono.org/pychrono/>`_ library.
"""

from wa_simulator._import import _import, _get_files
from wa_simulator import *

try:
    # Try to import chrono
    # Will through exception if failed
    import pychrono as chrono
    import pychrono.vehicle as veh
except Exception as e:
    raise e

for d in _get_files(__file__):
    _import(d, globals())


del _import, _get_files



"""
Wisconsin Autonomous - https://www.wisconsinautonomous.org

Copyright (c) 2021 wisconsinautonomous.org
All rights reserved.

Use of this source code is governed by a BSD-style license that can be found
in the LICENSE file at the top level of the repo
"""

# WA Simulator
from wa_simulator.vehicle import WAVehicle
from wa_simulator.core import WAVector, WAQuaternion
from wa_simulator.utils import _check_field, _load_json, get_wa_data_file, _WAStaticAttribute
from wa_simulator.chrono.utils import ChVector_to_WAVector, ChQuaternion_to_WAQuaternion, WAVector_to_ChVector, WAQuaternion_to_ChQuaternion, get_chrono_data_file, get_chrono_vehicle_data_file

# Chrono specific imports
import pychrono as chrono
import pychrono.vehicle as veh


def read_vehicle_model_file(filename: str) -> tuple:
    """Read a json specification file to get additional file names to be loaded into ChVehicle classes

    Will find the vehicle, powertrain and tire input files. The input files are other json files that are
    readable by the Chrono simulator to be used to create bodies attached to the vehicle.

    Args:
        filename (str): the json specification file with Vehicle, Powertrain and Tire input models

    Returns:
        tuple: returns each json specification file for the vehicle, powertrain and tire
    """
    j = _load_json(filename)

    # Validate json file
    _check_field(j, "Vehicle", field_type=dict)
    _check_field(j, "Powertrain", field_type=dict)
    _check_field(j, "Tire", field_type=dict)

    _check_field(j["Vehicle"], "Input File", field_type=str)
    _check_field(j["Powertrain"], "Input File", field_type=str)
    _check_field(j["Tire"], "Input File", field_type=str)

    # Extract the actual files
    vehicle_filename = veh.GetDataFile(j["Vehicle"]["Input File"])
    powertrain_filename = veh.GetDataFile(j["Powertrain"]["Input File"])
    tire_filename = veh.GetDataFile(j["Tire"]["Input File"])

    return vehicle_filename, powertrain_filename, tire_filename


def create_tire_from_json(tire_filename: str) -> veh.ChTire:
    """Creates a ChTire from a tire file

    .. info:
        Currently, only TMeasyTires and RigidTires are supported

    Args:
        tire_filename (str): the tire json specification file

    Returns:
        ChTire: the created tire

    Raises:
        TypeError: If the tire type is not recognized
    """
    j = _load_json(tire_filename)

    # Valide json file
    _check_field(j, "Type", value="Tire")
    _check_field(j, "Template", allowed_values=["TMeasyTire", "RigidTire", "Pac02Tire"])

    tire_type = j["Template"]
    if tire_type == "TMeasyTire":
        return veh.TMeasyTire(tire_filename)
    elif tire_type == "RigidTire":
        return veh.RigidTire(tire_filename)
    elif tire_type == "Pac02Tire":
        return veh.Pac02Tire(tire_filename)
    else:
        raise TypeError(f"'{tire_type} not a recognized tire type")


class WAChronoVehicle(WAVehicle):
    """Chrono vehicle wrapper

    Args:
        system (WAChronoSystem): the system used to run the simulation
        env (WAEnvironment): the environment with a terrain
        vehicle_inputs (WAVehicleInputs): the vehicle inputs
        filename (str): json file specification file
        init_loc (WAVector, optional): the inital location of the vehicle. Defaults to WAVector([0, 0, 0.5]).
        init_rot (WAQuaternion, optional): the initial orientation of the vehicle. Defaults to WAQuaternion([1, 0, 0, 0]).
    """

    # Global filenames for vehicle models
    _GO_KART_MODEL_FILE = "GoKart/GoKart.json"

    GO_KART_MODEL_FILE = _WAStaticAttribute('_GO_KART_MODEL_FILE', get_chrono_vehicle_data_file)

    def __init__(self, system: 'WAChronoSystem', vehicle_inputs: 'WAVehicleInputs', env: 'WAEnvironment', filename: str, init_loc: WAVector = WAVector([0, 0, 0.5]), init_rot: WAQuaternion = WAQuaternion([1, 0, 0, 0])):
        super().__init__(system, vehicle_inputs, get_wa_data_file("vehicles/GoKart/GoKart_KinematicBicycle.json"))

        # Get the filenames
        vehicle_file, powertrain_file, tire_file = read_vehicle_model_file(filename)

        # Create the vehicle
        vehicle = veh.WheeledVehicle(system._system, vehicle_file)

        # Initialize the vehicle
        init_loc = WAVector_to_ChVector(init_loc)
        init_rot = WAQuaternion_to_ChQuaternion(init_rot)
        vehicle.Initialize(chrono.ChCoordsysD(init_loc, init_rot))

        # Set the visualization components for the vehicle
        vehicle.SetChassisVisualizationType(veh.VisualizationType_MESH)
        vehicle.SetSuspensionVisualizationType(veh.VisualizationType_NONE)
        vehicle.SetSteeringVisualizationType(veh.VisualizationType_NONE)
        vehicle.SetWheelVisualizationType(veh.VisualizationType_MESH)

        # Create the powertrain
        # Assumes a SimplePowertrain
        powertrain = veh.SimplePowertrain(powertrain_file)
        vehicle.InitializePowertrain(powertrain)

        # Create and initialize the tires
        for axle in vehicle.GetAxles():
            tireL = create_tire_from_json(tire_file)
            vehicle.InitializeTire(
                tireL, axle.m_wheels[0], veh.VisualizationType_MESH)
            tireR = create_tire_from_json(tire_file)
            vehicle.InitializeTire(
                tireR, axle.m_wheels[1], veh.VisualizationType_MESH)

        self._vehicle = vehicle
        self._terrain = env._terrain

    def synchronize(self, time: float):
        d = veh.Inputs()
        d.m_steering = self._vehicle_inputs.steering
        d.m_throttle = self._vehicle_inputs.throttle
        d.m_braking = self._vehicle_inputs.braking

        self._vehicle.Synchronize(time, d, self._terrain)

    def advance(self, step: float):
        self._vehicle.Advance(step)

    def get_pos(self) -> WAVector:
        return ChVector_to_WAVector(self._vehicle.GetVehiclePos())

    def get_rot(self) -> WAQuaternion:
        return ChQuaternion_to_WAQuaternion(self._vehicle.GetVehicleRot())

    def get_pos_dt(self) -> WAVector:
        vel = self._vehicle.GetChassisBody().PointSpeedLocalToParent(
            self._vehicle.GetChassis().GetLocalDriverCoordsys().pos)
        vel = self._vehicle.GetChassisBody().GetRot().Rotate(vel)
        return ChVector_to_WAVector(vel)

    def get_rot_dt(self) -> WAQuaternion:
        return ChVector_to_WAVector(self._vehicle.GetChassisBody().GetWvel_loc())

    def get_pos_dtdt(self) -> WAVector:
        acc = self._vehicle.GetChassisBody().PointAccelerationLocalToParent(
            self._vehicle.GetChassis().GetLocalDriverCoordsys().pos)
        acc = self._vehicle.GetChassisBody().GetRot().Rotate(acc) - \
            self._vehicle.GetSystem().Get_G_acc()
        return ChVector_to_WAVector(acc)

    def get_rot_dtdt(self) -> WAQuaternion:
        return ChVector_to_WAVector(self._vehicle.GetChassisBody().GetWacc_loc())
